#
# x_check.pl: Program for identification of X-marker typing incompatibilities 
#             in pedigrees of arbitrary structure
	


$name="X_Check.pl";
$head="\n$name  (c) Irina Zorkoltseva 2005 Institute of Cytolgy and Genetics 
Novosibirsk Russia
Last revision:  March, 25, 2005\n";

#******** default parametrs ************************
$inp="pedfile.dat"; # default input pedigree data file
$mar="marker.dat";	# default input list of name of analyzed markers 
$err="x_check.err"; # default output file with errors information
$sep="\t";      	# default input separator
$error="";
#**************** check the comand line and reparametrization ************
if (@ARGV!=0)   {

	for ($i=0;$i<=@ARGV-1;$i++) 
	{
		$inp=$ARGV[$i+1] if ($ARGV[$i] =~ /-p/);
		$mar=$ARGV[$i+1] if ($ARGV[$i] =~ /-n/);
		$sep=$ARGV[$i+1] if ($ARGV[$i] =~ /-s/);
	}
 }

&PARAM;

&PED_READ; ## open and reading input file

open  MAC, "$mar" or die "ERROR: cannot open marker's list $mar: $!\n";

while (<MAC>) {
	next if (/^#/);
	next if (/^ *$/);
	chomp;
	$_="$_";
	$mset[$m]=$_;         ### @mset - array of marker's name
	$m++;
}
close (MAC, "$mar");

$ped=0;
$id=1;
$fat=2;
$mot=3;
$sex=4;


##for my $i ( 0..$#mset ) { print  "marker $mset[$i]\n"; }
print "*********** LEVEL 1 ERRORS ****************\n\n";
print ERR "*********** LEVEL 1 ERRORS ****************\n\n";


&CRTNYP;   # a partition  of pedigree into nuclear pedigrees

for $M ( 0..$#mset ) { ## marker's cykle 

$m=2*$M+5; # first marker's index in DAT

	for $p (1..$#NP ) {
	$error="";
       
if ($#{ $NP[$p] } == 1 ) {
	    $in= $NP[$p][0]-1;
	    & men($in) if ($DAT[$in][$m] ne "0" and $DAT[$in][$sex] eq "1");
	& error0_message if ($error ne ""); 
	 }
	else{
	$fa= $NP[$p][0]-1; # father's index 
	$mo= $NP[$p][1]-1; # mother's index

	& A if ($DAT[$fa][$m] ne "0" and $DAT[$mo][$m] ne "0");
	& SIBS if ($DAT[$fa][$m] eq "0" and $DAT[$mo][$m] eq "0");
	& C if ($DAT[$fa][$m] ne "0" and $DAT[$mo][$m] eq "0");
	& D if ($DAT[$fa][$m] eq "0" and $DAT[$mo][$m] ne "0");
	& error1_message if ($error ne ""); 
	    } ## end else	
			 } ## end NP cykle

 } ## end marker's cykle 

 & summary_message;	## output summary  information

close ERR;

####################################################################### 

sub PARAM {

open ERR, ">$err";
$date=localtime();
$param="\nDate: $date\n
 Input Files:
 Pedigree File: $inp
 Names file: $mar\n\n";   
print "$head$param";
print ERR "$head$param";

}

sub A { ## father and mother are genotyped both

$true=men($fa);
if ($true) { & D;}   ## error in father's genotype
   else {
    for my $i (2..$#{ $NP[$p] }) {
        $of= $NP[$p][$i]-1; # child's index 
        if ($DAT[$of][$m] ne "0") {        
          if ($DAT[$of][$sex] == 1) {  
	    $true=men($of);
	    & son_mo($of) unless($true)
 } 
          else {  
		& do_mo($of);
		& do_fa($of);
		if ($dm and $df) { & child_parent_diff($of);}
		}
	   }
}  
}  
}  # end sub A



sub C {        ## father's genotype is knowen

$true=men($fa);
unless ($true) {   ## not error in father's genotype
       for my $i (2..$#{ $NP[$p] }) {
        $of= $NP[$p][$i]-1; # child's index 
        if ($DAT[$of][$m]) { & do_fa($of) if ($DAT[$of][$sex] == 2) }  
}
}

&SIBS;

}       # end sub C


sub D {       ## father is untraited, mother is traited

    for my $i (2..$#{ $NP[$p] }) {
        $of= $NP[$p][$i]-1; # child's index 
        if ($DAT[$of][$m]) {        
          if ($DAT[$of][$sex] == 1) {  
	    $true=men($of);
	    & son_mo($of) unless($true)
 } 
          else {  
		& do_mo($of);
##		if ($dm) { & child_parent_diff($of);}
		}
	   }
}  
&SIBS;
}     # end sub D

sub SIBS {

@uniq=@uniq1=@uniq2=@uniq3=@uniq4=@female=@y=@list_het=@list_homo=@male=@list=@_;
my $n_homz=$er=0;
my $al="";
%seen=();
    for my $i (2..$#{ $NP[$p] }) {   
my $of= $NP[$p][$i]-1;

if ($DAT[$of][$m]){
push (@list,$DAT[$of][$m]);
push (@list,$DAT[$of][$m+1]);

if ($DAT[$of][$sex] == 1){
$true=men($of);
unless ($true) { push (@male,$DAT[$of][$m])}
}

if($DAT[$of][$sex] == 2){
	push (@female,$DAT[$of][$m],$DAT[$of][$m+1]);
	if ($DAT[$of][$m] eq $DAT[$of][$m+1]) {    ## homozygous female
	push (@list_homo,$DAT[$of][$m])}

else {       			## heterozygous female
	@y=($DAT[$of][$m],$DAT[$of][$m+1]);
	@y=sort(@y);
	$key="$y[0] $y[1]";
	push (@list_het,$key);
     }
                       	}

}
}  ## end cikle for offspring

## ERRORS TEST ##########################################

unless ($er) {
foreach $item (@list) {
    push (@uniq, $item) unless $seen{$item}++;   # uniq - set of child alleles
	}
if ($#uniq > 2){
$error="${error} ERROR: Children have more than 3 alleles \(@uniq\).\n";
	$no_er++;
	$er=1;
  }
}

unless ($er) {
%seen=();
foreach $item (@list_homo) {
    push (@uniq4, $item) unless $seen{$item}++;   # uniq4 - set of homozygotes female alleles
	}
if ($#uniq4 > 0){
	$error="${error} ERROR: Two or more different homozygotes (females) in sibship \n";
	$no_er++;
	$er=1;
}
}

unless ($er) {
if ($#uniq > 1 and $#uniq4> -1){
$error="${error} ERROR: Homozygous child and more than 2 alleles in sibship \(@uniq\).\n";
	$no_er++;
	$er=1;
  }
}


unless ($er) {
%seen=();
foreach $item (@list_het) {
    push (@uniq1, $item) unless $seen{$item}++;   # uniq1 - set of heterozygotes female genotypes
	}
if ($#uniq1 > 1){
$error="${error} ERROR: More then two different heterozygotes (females) in sibship \n";
	$no_er++;
	$er=1;
  }
}

unless ($er) {
%seen=();
foreach $item (@male) {
    push (@uniq2, $item) unless $seen{$item}++;   # uniq2 - set of male alleles
	}

if ($#uniq2 > 1){
$error="${error} ERROR: More then two different homozygotes (males) in sibship \n";
	$no_er++;
	$er=1;
  }
}

unless ($er) {
%seen=();
foreach $item (@female) {
    push (@uniq3, $item) unless $seen{$item}++;   # uniq3 - set of female alleles
	}
if (@uniq2 and $#uniq1 > 0  and $#uniq3 > 1) { ## test error 8
foreach $allel (@uniq2) {
$s=$ss=0;
    for  $i (2..$#{ $NP[$p] }) {   
my $of= $NP[$p][$i]-1;
if ($DAT[$of][$m] and $DAT[$of][$sex] == 2){
$ss++;
$s++ if ($allel eq $DAT[$of][$m] or $allel eq $DAT[$of][$m+1]); 
}
}
if ($s ==  $ss) {
$error="${error} ERROR: There are inconsistent in sibship \n";
	$no_er++;
	$er=1;

}
}
}
}
#######
} 	# end sub SIBS

sub child_parent_diff {
my $ind=$_[0];
my (@a,@b)=@_;
my %seen; # lookup table
my  @aonly;# answer
@a=($DAT[$ind][$m],$DAT[$ind][$m+1]);
@b=($DAT[$mo][$m],$DAT[$mo][$m+1],$DAT[$fa][$m]);
@seen{@b} = ();
foreach $item (@a) {
push(@aonly, $item) unless exists $seen{$item};
}
if ($#aonly >= 0){
$error="${error} ERROR: Child $DAT[$ind][$id] is consistent with each parent separately, but not as a pair.\n";
	$no_er++;
  }

}

sub son_mo {
my $ind=$_[0];
if  ($DAT[$ind][$m] ne $DAT[$mo][$m] and $DAT[$ind][$m] ne $DAT[$mo][$m+1]) {
             	$error="${error} ERROR: Child $DAT[$ind][$id] and Mother are inconsistent.\n";
		$no_er++;
  }
}

sub do_mo {
my $ind=$_[0];
$dm="true";
if ($DAT[$ind][$m] ne $DAT[$mo][$m] and $DAT[$ind][$m] ne $DAT[$mo][$m+1] and
     $DAT[$ind][$m+1] ne $DAT[$mo][$m] and $DAT[$ind][$m+1] ne $DAT[$mo][$m+1])
{	$error="${error} ERROR: Child $DAT[$ind][$id] and Mother are inconsistent.\n";
	$no_er++;
$dm="";
}
return $dm;
} 

sub do_fa {
my $ind=$_[0];
$df="true";
if ($DAT[$ind][$m] ne $DAT[$fa][$m] and $DAT[$ind][$m+1] ne $DAT[$fa][$m])
{	$error="${error} ERROR: Child $DAT[$ind][$id] and Father are inconsistent.\n";
	$no_er++;
$df="";
}
return $df;
} 


sub men {
my $ind=$_[0];
my $true=0;
if ($DAT[$ind][$m] ne $DAT[$ind][$m+1])  {
	$error="${error}ERROR: Per: $DAT[$ind][$id] is an X-linked male. He should be scored as homozygous.\n";
	$no_er++;
	$true=1;
}
return $true;
} 


sub error0_message {
my $MM=$M+1;
$pol="male";

$err_mes3="\n\#\#\# GENOTYPE ERROR: Pedigree $DAT[$in][$ped] Locus $MM Name $mset[$M] \#\#\#\#\n
 $error\n
 ORIGINAL SCORING:
 Per: $DAT[$in][$id] \($pol\): $DAT[$in][$m]\/$DAT[$in][$m+1]\n";
print ERR "$err_mes3";
##print "$err_mes3";

}
sub error1_message {

my $MM=$M+1;

$err_mes1="\n\#\#\# GENOTYPE ERROR: Pedigree $DAT[$fa][$ped] Locus $MM Name $mset[$M] \#\#\#\#\n
 $error\n
 ORIGINAL SCORING:
 Father $DAT[$fa][$id]: $DAT[$fa][$m]\/$DAT[$fa][$m+1]  Mother $DAT[$mo][$id]: $DAT[$mo][$m]\/$DAT[$mo][$m+1]\n";

print ERR "$err_mes1";    
##print "$err_mes1";    

for my $i (2..$#{ $NP[$p] }) { 
my $of=$NP[$p][$i]-1;
$pol="male" if ($DAT[$of][$sex] == 1);
$pol="female" if ($DAT[$of][$sex] == 2);
$err_mes2=" Child $DAT[$of][$id] \($pol\): $DAT[$of][$m]\/$DAT[$of][$m+1]\n";
print ERR "$err_mes2";
##print "$err_mes2";
}
} ##end error_message

sub summary_message {
$sum_m="\n
-------------------------------------------------
! Summary of Errors: By Pedigree                !
!                                               !
!                                               !
! Summary of Errors: By Marker                  !
-------------------------------------------------\n\n\n";

print ERR "$sum_m"; 
##print "$sum_m";

unless ($no_err) {
$str2=" X_Check has found $no_er inconsistencies in the pedigree data.\n
 See x_check_err file for details.
 After fixing the pedigree file rerun X_Check until
 there are no more errors.\n\n";
print ERR "$str2";
print  "$str2\n END OF PROGRAM\n";
} else {
$str1=" X_Check has found NO inconsistencies in the pedigree data.\n" ;
print ERR "$str1";
print "$str1\n END OF PROGRAM\n";
}

}   #end summary_message


sub CRTNYP {

# program for partition  of pedigree into nuclear pedigrees
$n_ped=1;

for (my $i=0; $i < $nids; $i++) {
next if ($nyp[$i]);
next unless ($DAT[$i][$fat]);
$nyp[$i]= $n_ped; 

## generation of massive massives

push @{ $NP[$n_ped] },"$DAT[$i][$fat]","$DAT[$i][$mot]","$DAT[$i][$id]"; ##  add fa, mo, first offspring to new NP

for (my $j=$i+1;$j<$nids;$j++) {
next  if ($nyp[$j]);
next unless ($DAT[$j][$fat]);

if ( $DAT[$j][$fat] eq $DAT[$i][$fat] and 
     $DAT[$j][$mot] eq $DAT[$i][$mot]) {
push @{ $NP[$n_ped] },"$DAT[$j][$id]"; ##  add offspring to NP
$nyp[$j]= $nyp[$i]; 
}
}

$n_ped++;
}



#**** OUTPUT NP - arrays of nuclear pedigree
#****open (ttt,">ttt");
#****for my $i ( 1..$#NP ) { print (ttt "NP $i: @{$NP[$i]}\n"); }
#****close ttt;

} #end CRTNYP

########################################################################
 
 sub PED_READ {
#************open and read the pedigree data******************
                                   
open(INP,"$inp") or die "ERROR: cannot open pedigree file $inped: $!\n"; 

while (<INP>) {
	next if /^#/;
	next if /^ *$/;
	chomp;
	@tmp=split/$sep/;
	push @DAT, [@tmp];

### print "@tmp\n";
}

$nids=scalar(@DAT);  ## it is equivalent $nids=@DAT;

}  ##end PED_READ

######################################################################
 

### error 8 !!!
### singlentone male !!!
### SUMMARY
